"""

Copyright (c) 2020 Ayoub Malek and Vanessa Sochat

This source code is licensed under the terms of the MIT license.
For a copy, see <https://opensource.org/licenses/MIT>.

"""

import os
import sys
from urlchecker.core import fileproc, urlproc
from urlchecker.core.whitelist import white_listed

def run_urlchecker(
    path,
    file_types,
    white_listed_files,
    white_listed_urls,
    white_listed_patterns,
    print_all,
    retry_count=2,
    timeout=5
):
    """
    Run the url checker given a path, a whitelist for each of url and file
    name paths or patterns, and a number of retries and timeouts.
    This provides a wrapper to check_files, which expects an already
    determined listing of files.

    Args:
        - path                   (str) : full path to the root folder to check.
        - print_all              (str) : control var for whether to print all checked file names or only the ones with urls.
        - white_listed_urls     (list) : list of white-listed urls.
        - white_listed_patterns (list) : list of white-listed patterns for urls.
        - white_listed_files    (list) : list of white-listed files and patterns for flies.
        - retry_count            (int) : number of retries on failed first check. Default=2.
        - timeout                (int) : timeout to use when waiting on check feedback. Default=5.

    Returns: dictionary with each of list of urls for "failed" and "passed."
    """
    if not os.path.exists(path):
        sys.exit("%s does not exist." % path)

    # get all file paths
    file_paths = fileproc.get_file_paths(
        base_path=path,
        file_types=file_types,
        white_listed_files=white_listed_files,
    )

    # check white listed list of files
    return check_files(
        file_paths=file_paths,
        print_all=print_all,
        white_listed_urls=white_listed_urls,
        white_listed_patterns=white_listed_patterns,
        retry_count=retry_count,
        timeout=timeout,
    )


def check_files(
    file_paths,
    print_all=True,
    white_listed_urls=None,
    white_listed_patterns=None,
    retry_count=2,
    timeout=5,
):
    """
    Check all urls extracted from all files in a given list.

    Args:
        - file_paths            (list) : list of file paths with urls to check (required).
        - print_all              (str) : control var for whether to print all checked file names or only the ones with urls.
        - white_listed_urls     (list) : list of white-listed urls.
        - white_listed_patterns (list) : list of white-listed url patterns.
        - retry_count            (int) : number of retries on failed first check. Default=1.
        - timeout                (int) : timeout to use when waiting on check feedback. Default=5.

    Returns:
        (list) check-results as a list of two lists (successfull checks, failed checks).
    """
    # init results list (first is success, second is issue)
    check_results = {"passed":[], "failed": []}

    # Allow for user to skip specifying white listed options
    white_listed_urls = white_listed_urls or []
    white_listed_patterns = white_listed_patterns or []

    # loop files
    for file_name in file_paths:

        # collect links from each file (unique=True is set)
        urls = fileproc.collect_links_from_file(file_name)

        # eliminate white listed urls and white listed white listed patterns
        if white_listed_urls or white_listed_patterns:
            urls = [
                url
                for url in urls
                if not white_listed(url, white_listed_urls, white_listed_patterns)
            ]

        # if some links are found, check them
        if urls:
            print("\n", file_name, "\n", "-" * len(file_name))
            urlproc.check_urls(urls, check_results, retry_count, timeout)

        # if no urls are found, mention it if required
        else:
            if print_all:
                print("\n", file_name, "\n", "-" * len(file_name))
                print("No urls found.")

    return check_results



"""

Copyright (c) 2020 Ayoub Malek and Vanessa Sochat

This source code is licensed under the terms of the MIT license.  
For a copy, see <https://opensource.org/licenses/MIT>.

"""

import logging

def print_failure(message):
    """
    Given a message string, print as a failure in red.

    Parameters:
      - message: the message to print in red (indicating failure).
    """
    print("\x1b[31m" + message + "\x1b[0m")


def print_success(message):
    """
    Given a message string, print as a success in green.

    Parameters:
      - message: the message to print in green (indicating success).
    """
    print("\x1b[32m" + message + "\x1b[0m")


def get_logger(name='urlchecker', level=logging.INFO):
    """
    Get a default logger for the urlchecker library, meaning
    that we use name "urlchecker" and use the default logging
    level INFO
 
    Parameters:
      - name: the name for the logger (defaults to urlchecker)
      - level: the logging.<level> to set with setLevel()

    Returns: logging logger
    """

    logger = logging.getLogger(name)
    logger.setLevel(level)

    # Stream handler
    ch = logging.StreamHandler()
    ch.setLevel(logging.ERROR)

    # formatting
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')    
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger



"""

Copyright (c) 2020 Ayoub Malek and Vanessa Sochat

This source code is licensed under the terms of the MIT license.  
For a copy, see <https://opensource.org/licenses/MIT>.

"""

from urlchecker.version import __version__



"""
client/github.py: entrypoint for interaction with a GitHub repostiory.
Copyright (c) 2020 Ayoub Malek and Vanessa Sochat
"""

import re
import os
import sys
import logging

from urlchecker.main.github import clone_repo, delete_repo
from urlchecker.core.fileproc import remove_empty, save_results
from urlchecker.core.check import run_urlchecker
from urlchecker.logger import print_success, print_failure

logger = logging.getLogger('urlchecker')


def main(args, extra):
    """
    main entrypoint for running a check. We expect an args object with
    arguments from the main client. From here we determine the path
    to parse (or GitHub url to clone) and call the main check function
    under main/check.py

    Parameters:
      - args: the argparse ArgParser with parsed args
      - extra: extra arguments not handled by the parser
    """
    path = args.path

    # Case 1: specify present working directory
    if not path or path == ".":
        path = os.getcwd()
        logging.debug("Path specified as present working directory, %s" % path)

    # Case 2: We need to clone
    elif re.search("^(git@|http)", path):
        logging.debug("Repository url %s detected, attempting clone" % path)
        path = clone_repo(path, branch=args.branch)

    # Add subfolder to path
    if args.subfolder:
        path = os.path.join(path, args.subfolder)

    # By the time we get here, a path must exist
    if not os.path.exists(path):
        sys.exit("Error %s does not exist." % path)

    # Parse file types, and white listed urls and files (includes absolute and patterns)
    file_types = args.file_types.split(",")
    white_listed_urls = remove_empty(args.white_listed_urls.split(","))
    white_listed_patterns = remove_empty(args.white_listed_patterns.split(","))
    white_listed_files = remove_empty(args.white_listed_files.split(","))

    # Alert user about settings
    print("  original path: %s" % args.path)
    print("     final path: %s" % path)
    print("      subfolder: %s" % args.subfolder)
    print("         branch: %s" % args.branch)
    print("        cleanup: %s" % args.cleanup)
    print("     file types: %s" % file_types)
    print("      print all: %s" % (not args.no_print))
    print(" url whitetlist: %s" % white_listed_urls)
    print("   url patterns: %s" % white_listed_patterns)
    print("  file patterns: %s" % white_listed_files)
    print("     force pass: %s" % args.force_pass)
    print("    retry count: %s" % args.retry_count)
    print("           save: %s" % args.save)
    print("        timeout: %s" % args.timeout)

    # Run checks, get lookup of results and fails
    check_results = run_urlchecker(path=path,
                                   file_types=file_types,
                                   white_listed_files=white_listed_files,
                                   white_listed_urls=white_listed_urls,
                                   white_listed_patterns=white_listed_patterns,
                                   print_all=not args.no_print,
                                   retry_count=args.retry_count,
                                   timeout=args.timeout)

    # save results to flie, if save indicated
    if args.save:
        save_results(check_results, args.save)

    # delete repo when done, if requested
    if args.cleanup:
        logger.info("Cleaning up %s..." % path)
        delete_repo(path)

    # Case 1: We didn't find any urls to check
    if not check_results['failed'] and not check_results['passed']:
        print("\n\nDone. No urls were collected.")
        sys.exit(0)

    # Case 2: We had errors, but force pass is True
    elif args.force_pass and check_results['failed']:
        print("\n\nDone. The following urls did not pass:")
        for failed_url in check_results['failed']:
            print_failure(failed_url)
        sys.exit(1)

    else:
        print("\n\nDone. All URLS passed.")
        sys.exit(0)



#!/usr/bin/python

"""

Copyright (c) 2020 Ayoub Malek and Vanessa Sochat

This source code is licensed under the terms of the MIT license.
For a copy, see <https://opensource.org/licenses/MIT>.

"""


import os
import sys
import argparse
import urlchecker
import logging

def get_parser():
    parser = argparse.ArgumentParser(description="urlchecker python")

    # Global Variables
    parser.add_argument(
        "--debug",
        dest="debug",
        help="use verbose logging to debug.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--version",
        dest="version",
        help="suppress additional output.",
        default=False,
        action="store_true",
    )

    subparsers = parser.add_subparsers(
        help="urlchecker python actions",
        title="actions",
        description='actions for urlchecker',
        dest="command",
    )

    # print version and exit
    subparsers.add_parser(
        "version", help="show software version"  # pylint: disable=unused-variable
    )

    # main check entrypoint
    check = subparsers.add_parser(
        "check", help="check urls in static files (documentation or code)"
    )

    # supports a clone URL or a path
    check.add_argument(
        "path",
        help="the local path or GitHub repository to clone and check",
    )

    check.add_argument(
        "-b",
        "--branch",
        help="if cloning, specify a branch to use (defaults to master)",
        default="master",
    )

    check.add_argument(
        "--subfolder",
        help="relative subfolder path within path (if not specified, we use root)",
    )

    check.add_argument(
        "--cleanup",
        help="remove root folder after checking (defaults to False, no cleaup)",
        default=False,
        action="store_true",
    )

    check.add_argument(
        "--force-pass",
        help="force successful pass (return code 0) regardless of result",
        default=False,
        action="store_true",
    )

    check.add_argument(
        "--no-print",
        help="Skip printing results to the screen (defaults to printing to console).",
        default=False,
        action="store_true",
    )

    check.add_argument(
        "--file-types",
        dest="file_types",
        help="comma separated list of file extensions to check (defaults to .md,.py)",
        default=".md,.py",
    )

# White listing

    check.add_argument(
        "--white-listed-urls",
        help="comma separated list of white listed urls (no spaces)",
        default="",
    )

    check.add_argument(
        "--white-listed-patterns",
        help="comma separated list of white listed patterns for urls (no spaces)",
        default="",
    )

    check.add_argument(
        "--white-listed-files",
        help="comma separated list of white listed files and patterns for files (no spaces)",
        default="",
    )

# Saving

    check.add_argument(
        "--save",
        help="Path toa csv file to save results to.",
        default=None,
    )

# Timeouts

    check.add_argument(
        "--retry-count",
        help="retry count upon failure (defaults to 2, one retry).",
        type=int,
        default=2,
    )

    check.add_argument(
        "--timeout",
        help="timeout (seconds) to provide to the requests library (defaults to 5)",
        type=int,
        default=5,
    )

    return parser


def main():
    """main is the entrypoint urlchecker-python.
    """

    parser = get_parser()

    def help(return_code=0):
        """print help, including the software version and active client 
           and exit with return code.
        """

        version = urlchecker.__version__

        print("\nurlchecker python v%s" % version)
        parser.print_help()
        sys.exit(return_code)

    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()

    # If an error occurs while parsing the arguments, the interpreter will exit with value 2
    args, extra = parser.parse_known_args()

    # Show the version and exit
    if args.command == "version" or args.version:
        print(urlchecker.__version__)
        sys.exit(0)

    if args.command == "check":
        from .check import main
    else:
        print("Unsupported command %s" % args.command)
        sys.exit(0)
    
    # Pass on to the correct parser
    return_code = 0
    try:
        main(args=args, extra=extra)
        sys.exit(return_code)
    except UnboundLocalError:
        return_code = 1

    help(return_code)


if __name__ == "__main__":
    main()



"""

Copyright (c) 2020 Ayoub Malek and Vanessa Sochat

This source code is licensed under the terms of the MIT license.  
For a copy, see <https://opensource.org/licenses/MIT>.

"""

import csv
import os
import re
import sys
from urlchecker.core import urlmarker


def check_file_type(file_path, file_types):
    """
    Check file type to assert that only file with certain predefined extensions
    are checked.

    Args:
        - file_path   (str) : path to file.
        - file_types (list) : list of file extensions to accept.

    Returns:
        (bool) true if file type is supported else false.
    """
    ftype = "." + file_path.split(".")[-1]
    if ftype in file_types:
        return True

    # default return
    return False


def include_file(file_path, white_list_patterns):
    """
    Check a file path for inclusion based on an OR regular expression.
    The user is currently not notified if a file is marked for removal.

    Args:
        - file_path            (str) : a file path to check if should be included.
        - white_list_patterns (list) : list of patterns to whitelist (include).

    Returns:
        (bool) boolean indicating if the URL should be white listed (included).
    """
    # No white listed patterns, all files are included
    if not white_list_patterns:
        return True

    # Return False (don't include) if the patterns match
    regexp = "(%s)" % "|".join(white_list_patterns)
    return not re.search(regexp, file_path)


def get_file_paths(base_path, file_types, white_listed_files=None):
    """
    Get path to all files under a give directory and its subfolders.

    Args:
        - base_path           (str) : base path.
        - file_types         (list) : list of file extensions to accept.
        - white_listed_files (list) : list of files or patterns to white list

    Returns:
        (list) list of file paths.
    """
    white_listed_files = white_listed_files or []

    # init paths
    file_paths = []

    # walk folders and colect file paths
    for root, directory, files in os.walk(base_path):
        file_paths += [
            os.path.join(root, file)
            for file in files
            if os.path.isfile(os.path.join(root, file))
            and check_file_type(file, file_types)
            and include_file(os.path.join(root, file), white_listed_files)
        ]
    return file_paths


def collect_links_from_file(file_path, unique=True):
    """
    Collect all links in a file.

    Args:
        - file_path   (str) : path to file.

    Returns:
        (list) list of links/ urls in a file.
    """
    # read file content
    with open(file_path, "r") as file:
        content = file.read()

    # get and filter urls
    urls = re.findall(urlmarker.URL_REGEX, content)
    urls = [url.strip() for url in urls if url.strip().startswith("http")]
    urls = [url.strip("\\n") if url.endswith("\\n") else url for url in urls]

    # Do we only want unique links?
    if unique:
        return list(set(urls))

    return urls


def remove_empty(file_list):
    """
    Given a file list, return only those that aren't empty string or None.

    Args:
        - file_list (list): a list of files to remove None or empty string from.

    Returns:
        (list) list of (non None or empty string) contents.
    """
    return [x for x in file_list if x not in ["", None]]


def save_results(check_results, file_path, sep=",", header=None):
    """
    Given a check_results dictionary, a dict with "failed" and "passed" keys (
    or more generally, keys to indicate some status), save a csv
    file that has header with URL,RESULT that indicates, for each url,
    a pass or failure. If the directory of the file path doesn't exist, exit
    on error.

    Args:
        - check_results (dict): the check results dictionary with passed/failed
        - file_path (str): the file path (.csv) to save to.
        - sep (str): the separate to use (defaults to comma)
        - header (list): if not provided, will save URL,RESULT

    Returns:
        (str) file_path: a newly saved csv with the results
    """
    # Ensure that the directory exists
    file_path = os.path.abspath(file_path)
    dirname = os.path.dirname(file_path)

    if not os.path.exists(dirname):
        sys.exit("%s does not exist, cannot save %s there." %(dirname, file_path))

    # Ensure the header is provided and correct (length 2)
    if not header:
        header = ["URL", "RESULT"]

    if len(header) != 2:
        sys.exit("Header must be length 2 to match size of data.")

    print("Saving results to %s" % file_path)

    # Write to file after header row
    with open(file_path, mode='w') as fd:
        writer = csv.writer(fd, delimiter=sep, quotechar='"', quoting=csv.QUOTE_MINIMAL)
        writer.writerow(header)
        for result, items in check_results.items():
            [writer.writerow([item, result]) for item in items];

    return file_path



"""

Copyright (c) 2020 Ayoub Malek and Vanessa Sochat

This source code is licensed under the terms of the MIT license.  
For a copy, see <https://opensource.org/licenses/MIT>.

"""

import os
import time
import random
import requests
from urlchecker.core import urlmarker
from urlchecker.logger import print_success, print_failure

def record_response(url, response, check_results):
    """
    Record response status of an input url. This function is run after success,
    or at the end of retry to record the final response.

    Args:
        - url          (str) : url text.
        - response    (list) : request response from the url request.
        - check_results (list) : list of lists, success appended to 0, failure to 1.
    """
    # response of None indicates a failure
    if not response:
        check_results["failed"].append(url)

    # success
    elif response.status_code == 200:
        check_results["passed"].append(url)

    # Any other error
    else:
        check_results["failed"].append(url)


def check_response_status_code(url, response):
    """
    Check response status of an input url. Returns a boolean
    to indicate if retry is needed.

    Args:
        - url          (str) : url text.
        - response    (list) : request response from the url request.

    Returns:
        (bool) boolean to indicate if retry is needed (True) or not (False)
    """
    # Case 1: response is None indicating triggered error
    if not response:
        print_failure(url)
        return True

    # Case 2: success! Retry is not needed.
    if response.status_code == 200:
        print_success(url)
        return False

    # Case 3: failure of some kind
    print_failure(url)
    return True


def get_user_agent():
    """
    Return a randomly chosen user agent for requests

    Returns:
        user agent string to include with User-Agent.
    """
    agents = [
        (
            "Mozilla/5.0 (X11; Linux x86_64) "
            "AppleWebKit/537.36 (KHTML, like Gecko) "
            "Chrome/57.0.2987.110 "
            "Safari/537.36"
        ),  # chrome
        (
            "Mozilla/5.0 (X11; Linux x86_64) "
            "AppleWebKit/537.36 (KHTML, like Gecko) "
            "Chrome/61.0.3163.79 "
            "Safari/537.36"
        ),  # chrome
        (
            "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:55.0) "
            "Gecko/20100101 "
            "Firefox/55.0"
        ),  # firefox
        (
            "Mozilla/5.0 (X11; Linux x86_64) "
            "AppleWebKit/537.36 (KHTML, like Gecko) "
            "Chrome/61.0.3163.91 "
            "Safari/537.36"
        ),  # chrome
        (
            "Mozilla/5.0 (X11; Linux x86_64) "
            "AppleWebKit/537.36 (KHTML, like Gecko) "
            "Chrome/62.0.3202.89 "
            "Safari/537.36"
        ),  # chrome
        (
            "Mozilla/5.0 (X11; Linux x86_64) "
            "AppleWebKit/537.36 (KHTML, like Gecko) "
            "Chrome/63.0.3239.108 "
            "Safari/537.36"
        ),  # chrome
    ]
    return random.choice(agents)


def check_urls(urls, check_results, retry_count=1, timeout=5):
    """
    Check urls extracted from a certain file and print the checks results.

    Args:
        - urls          (list) : list of urls to check.
        - check_results (list) : a list containing a list of succesfully checked links and errenous links.
        - retry_count    (int) : a number of retries to issue (defaults to 1, no retry).
        - timeout        (int) : a timeout in seconds for blocking operations like the connection attempt.
    """
    # init seen urls list
    seen = set()

    # Some sites will return 403 if it's not a "human" user agent
    user_agent = get_user_agent()
    headers = {"User-Agent": user_agent}

    # check links
    for url in [url for url in urls if "http" in url]:

        # init do retrails and retrails counts
        do_retry = True
        rcount = retry_count

        # we will double the time for retry each time
        retry_seconds = 2

        # With retry, increase timeout by a second
        pause = timeout

        # get url termination
        url_termination = "." + os.path.basename(url).split(".")[-1]

        # No need to test the same URL twice
        if url in seen:
            continue

        seen.add(url)
        while rcount > 0 and do_retry:
            response = None
            try:
                response = requests.get(url, timeout=pause, headers=headers)

            except requests.exceptions.Timeout as e:
                print(e)

            except requests.exceptions.ConnectionError as e:
                print(e)

            except Exception as e:
                print(e)

            # decrement retrials count
            rcount -= 1

            # Break from the loop if we have success, update user
            do_retry = check_response_status_code(url, response)

            # If we try again, pause for retry seconds and update retry seconds
            if rcount > 0 and do_retry:
                # keep this only for debugging
                # print("Retry n° %s for %s, with timeout of %s seconds." % (retry_count - rcount, url, pause))
                time.sleep(retry_seconds)
                retry_seconds = retry_seconds * 2
                pause += 1

        # When we break from while, we record final response
        record_response(url, response, check_results)



"""

Copyright (c) 2020 Ayoub Malek and Vanessa Sochat

This source code is licensed under the terms of the MIT license.
For a copy, see <https://opensource.org/licenses/MIT>.

"""

def white_listed(url, white_listed_urls, white_listed_patterns):
    """
    Check if link is in the white listed URLs or patterns to ignore.

    Args:
        - url                    (str) : link to check.
        - white_listed_urls     (list) : list of white-listed urls.
        - white_listed_patterns (list) : list of white-listed patterns.

    Returns:
        (bool) boolean for whether link is white-listed or not.
    """
    # check white listed urls
    if url in white_listed_urls:
        return True

    # check white listed patterns
    i = 0
    while i < len(white_listed_patterns):
        if white_listed_patterns[i] in url:
            return True
        i += 1

    # default return
    return False



"""

Copyright (c) 2020 Ayoub Malek and Vanessa Sochat

This source code is licensed under the terms of the MIT license.  
For a copy, see <https://opensource.org/licenses/MIT>.

"""

import os
import sys
import subprocess
from urlchecker.core import urlproc
from urlchecker.core import fileproc


def clone_repo(git_path, branch="master"):
    """
    Clone and name a git repository.

    Args:
        - git_path (str) : https path to git repository.
        - branch   (str) : name of the brankch to use. Default="master"

    Returns:
        (str) base path of the cloned git repository.
    """
    base_path = os.path.basename(git_path)
    result = subprocess.run(
        ["git", "clone", "-b", branch, git_path, base_path],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    if result.returncode != 0:
        sys.exit("Issue with cloning branch %s of %s" % (branch, git_path))

    return base_path


def delete_repo(base_path):
    """
    Delete repository.

    Args:
        - base_path (str) : base path of the cloned git repository.

    Returns:
        (str) message/ code describing whether the operation was successfully excuted.
    """
    # clone repo
    result = subprocess.run(
        ["rm", "-R", "-f", base_path], stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    return result.returncode



def get_branch(default="master"):
    """
    Derive the selected branch. We first look to the environment variable
    for INPUT_BRANCH, meaning that the user set the branch variable. If
    that is unset we parse GITHUB_REF. If both of those are unset,
    then we default to default (master).

    Returns:
        (str) the branch found in the environment, otherwise master.
    """
    # First check goes to use setting in action
    branch = os.getenv("INPUT_BRANCH")
    if branch:
        return branch

    # Second check is for GITHUB_REF
    branch = os.getenv("GITHUB_REF")
    if branch:
        branch = branch.replace("refs/heads/", "")
        return branch
    return default
